//
// load_textdata.h - static text loading module to access data
//                   which was linked into the executable, such
//                   as the assembly routines used in code generation,
//                   license text, etc.
//
// Copyright (C) 2017,2018,2019,2020,2021  John Gatewood Ham
//
// This file is part of ecma55.
//
// ecma55 is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2
// as published by the Free Software Foundation.
//
// ecma55 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ecma55.  If not, see <http://www.gnu.org/licenses/>.
//
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <search.h>
#include <stdbool.h>
#include <ctype.h>
#ifdef __PIE__
#define __USE_GNU
#include <dlfcn.h>
#endif
#include "load_textdata.h"
#include "textdata.h"
#include "error_messages.h"
#include "globals.h"

// everybody sets __GNUC__, so have to do it this ugly way
// optimization must be disabled for gcc
#if defined(__clang__)
#define WART
#elif defined(__TINYC__)
#define WART
#elif defined(__PCC__)
#define WART
#elif defined(__GNUC__)
#define WART __attribute__((optimize("-O0")))
#else
#define WART
#endif

volatile struct ilist *item_list=NULL;     // pointer to vector of assembly text pointers
static volatile uint64_t item_count=0;     // number of elements in item_list[] vector

#ifdef __PIE__
static void pie_setpointers(void) WART;
#else
static void exe_setpointers(void) WART;
#endif

#ifdef __PIE__
static void pie_setpointers(void) {
  void *hndl;     // pointer to dlopen()'d self
  char *errmsg;   // pointer to error message from dlerror()

  hndl = dlopen (NULL, RTLD_NOW | RTLD_NODELETE | RTLD_GLOBAL);
  if (!hndl) {
    fprintf(stderr, emsg[212], dlerror());
    exit (EXIT_FAILURE);
  };
  item_list = (struct ilist *)dlsym (hndl, "textchunk_ndex");
  errmsg = dlerror();
  if (errmsg) {
    fprintf(stderr, emsg[213], "textchunk_ndex", emsg);
    exit (EXIT_FAILURE);
  }
  item_count = *((uint64_t *)dlsym (hndl, "textchunk_count"));
  errmsg = dlerror();
  if (errmsg) {
    fprintf(stderr, emsg[213], "textchunk_count", emsg);
    exit (EXIT_FAILURE);
  }
  dlclose (hndl);
  return;
}
#else
static void exe_setpointers(void) {
  extern volatile void *textchunk_ndex;          // pointer to start of vector of assembly text pointers
  extern volatile uint64_t textchunk_count;      // number of elements in vector

  item_list = (struct ilist *)(&textchunk_ndex);
  item_count = (uint64_t)textchunk_count;
  return;
}
#endif

void initialize_item_list(void) {
#ifdef __PIE__
  pie_setpointers();
#else
  exe_setpointers();
#endif
  return;
}

#ifdef UNIT_TEST
static char errbuffer[4096] = { 0 };    // error message buffer

static int compare(const void *left, const void *right);
static void listall(void);
static bool lookup(const char *key, struct ilist **ptr);
static bool test_direct(unsigned int key, const char *realname, const char *fullpath);

static void listall(void) {
  printf("item_list=%p\n", (void *)item_list);
  fflush(stdout);
  printf("entries=%" PRIu64 "\n", item_count);
  fflush(stdout);
  for (unsigned int i = 0U; i < item_count; ++i) {
    char *outgoing_data = (char *)item_list[i].data;
    ssize_t howmanyleft = item_list[i].datalength,
            write_count = 0;

    fprintf(stdout, "chunk #%u:\n-----< begin '%s' >-----\n",
            i, item_list[i].fname);
    fflush(stdout);
    do {
      write_count = write(STDOUT_FILENO, (void *)(outgoing_data),
                          (size_t)howmanyleft);
      if (write_count<0) {
        switch (errno) {
          case EPIPE:
            fputs("Program terminated due to pipe closure.\n", stderr);
            return;
          default:
            ;// FALLTHRU
        }
        perror("Bad write");
        exit(EXIT_FAILURE);
      }
      howmanyleft -= write_count;
      outgoing_data += write_count;
    } while (howmanyleft > 0);
    fflush(stdout);
    if (howmanyleft < 0) {
      perror("Bad write\n");
      exit(EXIT_FAILURE);
    }
    fprintf(stdout, "------< end '%s' >------\n", item_list[i].fname);
    fflush(stdout);
  }
  return;
}

static int compare(const void *left, const void *right) {
  char **x = (char **)right;
  return strcmp((const char *)left, (const char *)(*x));
}

static bool lookup(const char *key, struct ilist **ptr) {
  *ptr=(struct ilist *)lsearch(key,
                               (void * )item_list,
                               (size_t *)&item_count,
                               sizeof(struct ilist),
                               compare);
  return (*ptr?true:false);
}

static bool test_lookups(void) {
  for (unsigned int i = 0U; i < item_count; ++i) {
    struct ilist *ptr;

    if (!lookup(item_list[i].fname, &ptr)) {
      fprintf(stderr, "Unable to find '%s'\n", item_list[i].fname);
      exit(EXIT_FAILURE);
    }
    if (item_list[i].datalength != ptr->datalength) {
      fprintf(stderr, "Expected %" PRIu64 " bytes, got %" PRIu64 " bytes\n",
              item_list[i].datalength, ptr->datalength);
      return false;
    }
    if (memcmp(ptr->data, item_list[i].data, ptr->datalength)) {
      fprintf(stderr, "For '%s' with %" PRIu64 " bytes\n",
              item_list[i].fname, item_list[i].datalength);
      fprintf(stderr, "Wanted:\n");
      hexdump(item_list[i].data, item_list[i].datalength, stderr);
      fputs("Got:\n", stderr);
      hexdump(ptr->data, ptr->datalength, stderr);
      return false;
    }
  }
  return true;
}

static bool test_direct(unsigned int key,
                        const char *realname,
                        const char *fullpath) {
  FILE *f = NULL;               // file handle
  size_t reallen,               // length of file in bytes
         totalsofar;            // current total number of bytes read
  unsigned char *realdata;      // buffer to contains the file contents
  struct stat sb;               // stat buffer

  if (stat(fullpath, &sb) < 0) {
    fprintf(stderr,"Cannot stat '%s'\n",fullpath);
    return false;
  }
  if (!((S_ISREG(sb.st_mode) || S_ISLNK(sb.st_mode)))) {
    fprintf(stderr, "File '%s' is not a regular file\n", fullpath);
    return false;
  }
  if ((f = fopen(fullpath, "rb")) == NULL) {
    sprintf(errbuffer, "Unable to open file '%s' for input", fullpath);
    perror(errbuffer);
    return false;
  }
  reallen = sb.st_size;
  if ((realdata = (unsigned char *)malloc(sizeof(char) * (reallen + 1U))) == NULL) {
    fclose(f);
    f = NULL;
    perror("Unable to allocate memory");
    return false;
  }
  realdata[0] = 0;
  totalsofar = 0;
  do {
    size_t howmany = fread(realdata, 1, reallen - totalsofar, f);
    if (feof(f)) {
      free(realdata);
      realdata = NULL;
      sprintf(errbuffer, "fread() EOF for input file '%s'", fullpath);
      perror(errbuffer);
      fclose(f);
      return NULL;
    }
    if (ferror(f)) {
      free(realdata);
      realdata = NULL;
      sprintf(errbuffer, "fread() Error for input file '%s'", fullpath);
      perror(errbuffer);
      fclose(f);
      return NULL;
    }
    totalsofar += howmany;
  } while (totalsofar < reallen);
  fclose(f);
  f = NULL;
  realdata[reallen] = 0;     // guarantee termination
  if (strcmp(item_list[key].fname,realname)) {
    free(realdata);
    realdata = NULL;
    fprintf(stderr,"expected '%s', got '%s'\n",realname,item_list[key].fname);
    return false;
  }
  if (item_list[key].datalength!=reallen) {
    free(realdata);
    realdata = NULL;
    fprintf(stderr,"expected %" PRIu64 ", got %" PRIu64 "\n",reallen,item_list[key].datalength);
    return false;
  }
  if (memcmp(item_list[key].data,realdata,reallen)) {
    fprintf(stderr,"data mismatch!\n"
                   "For '%s' with %" PRIu64 " bytes\n",
            item_list[key].fname, item_list[key].datalength);
    fputs("Wanted:\n", stderr);
    hexdump(item_list[key].data, item_list[key].datalength, stderr);
    fputs("Got:\n", stderr);
    hexdump(realdata, reallen, stderr);
    free(realdata);
    realdata = NULL;
    return false;
  }
  free(realdata);
  realdata = NULL;
  return true;
}

int main(void) {
  int retval=EXIT_SUCCESS;

  initialize_item_list();
  if (!test_lookups()) {
    puts("Lookup tests failed.");
    retval=EXIT_FAILURE;
  } else {
    puts("Lookup tests passed.");
    if (!test_direct(floor_asmcode,"floor.s","AMD64/floor.s")) {
      puts("test_direct failed.");
      retval=EXIT_FAILURE;
    } else {
      puts("test_direct passed.");
    }
  }
  if (retval!=EXIT_SUCCESS)
    listall();
  return retval;
}
#endif
